@@ -23,10 +23,10 @@ import sys
 import tempfile
 
 import pyficache
-
 from xdis import IS_PYPY, PYTHON_VERSION_TRIPLE, load_module
 from xdis.disasm import disassemble_file
 from xdis.version_info import version_tuple_to_str
+
 from trepan.client import run
 from trepan.clifns import whence_file
 from trepan.debugger import Trepan
@@ -90,10 +90,10 @@ def main(dbg=None, sys_argv=list(sys.argv)):
             mainpyfile = whence_file(mainpyfile)
             is_readable = readable(mainpyfile)
             if is_readable is None:
-                print(f"{__title__}: Python script file '{mainpyfile}' does not exist")
+                print(f"{__title__}: Python script file '{mainpyfile}' does not exist.")
                 sys.exit(1)
             elif not is_readable:
-                print(f"{__title__}: Can't read Python script file '{mainpyfile}'")
+                print(f"{__title__}: Can't read Python script file '{mainpyfile}.'")
                 sys.exit(1)
                 return
 
@@ -109,15 +109,22 @@ def main(dbg=None, sys_argv=list(sys.argv)):
                     _,
                 ) = load_module(mainpyfile, code_objects=None, fast_load=False)
                 if is_pypy != IS_PYPY:
-                    print(f"Bytecode is pypy {is_pypy}, but we are {IS_PYPY}.")
+                    bytecode_pypy = "" if is_pypy else "not "
+                    running_pypy = "" if IS_PYPY else "not "
+                    print(
+                        f"Bytecode is for Version {version_tuple_to_str(end=2)} {bytecode_pypy}running PyPY, but we are {running_pypy}running PyPy."
+                    )
                     print("For a cross-version debugger, use trepan-xpy with x-python.")
                     sys.exit(2)
                 if python_version[:2] != PYTHON_VERSION_TRIPLE[:2]:
+                    bytecode_pypy = " PyPy" if is_pypy else ""
                     print(
-                        f"Bytecode is for version {version_tuple_to_str(python_version)} but we are "
-                        f"version {version_tuple_to_str()}."
+                        f"Bytecode is for version {version_tuple_to_str(python_version, end=2)}{bytecode_pypy}, but we are "
+                        f"version {version_tuple_to_str(end=2)}{bytecode_pypy}."
+                    )
+                    print(
+                        "For a cross-version debugger, trepan-xpy with x-python might work."
                     )
-                    print("For a cross-version debugger, trepan-xpy with x-python might work.")
                     sys.exit(2)
 
                 py_file = co.co_filename
@@ -141,22 +148,36 @@ def main(dbg=None, sys_argv=list(sys.argv)):
                     raise IOError(
                         f"Python file name embedded in code {try_file} not found"
                     )
+            except ImportError as e:
+                print(str(e))
+                sys.exit(3)
             except IOError:
                 decompiler = "uncompyle6"
+                no_decompiler = False
                 try:
                     if (3, 7) <= PYTHON_VERSION_TRIPLE <= (3, 8):
-                        from uncompyle6 import decompile_file
-                    else:
                         from decompyle3 import decompile_file
+                    elif PYTHON_VERSION_TRIPLE >= (3, 9):
+                        decompiler = "No decompiler found"
+                        no_decompiler = True
+                        raise ImportError
+                    else:
+                        from uncompyle6 import decompile_file
 
                         decompiler = "decompyle3"
                 except ImportError:
-                    print(
-                        "%s: Compiled python file '%s', but %s not found"
-                        % (__title__, mainpyfile, decompiler),
-                        file=sys.stderr,
-                    )
-                    sys.exit(1)
+                    if PYTHON_VERSION_TRIPLE >= (3, 9):
+                        print(
+                            "%s: Decompiler not available for %s."
+                            % (__title__, version_tuple_to_str()),
+                            file=sys.stderr,
+                        )
+                    else:
+                        print(
+                            "%s: Compiled python file '%s', but %s not found"
+                            % (__title__, mainpyfile, decompiler),
+                            file=sys.stderr,
+                        )
 
                 short_name = osp.basename(mainpyfile).strip(".pyc")
                 fd = tempfile.NamedTemporaryFile(
@@ -179,16 +200,26 @@ def main(dbg=None, sys_argv=list(sys.argv)):
 
                 # from io import StringIO
                 # linemap_io = StringIO()
-                try:
-                    decompile_file(mainpyfile, fd.file, mapstream=fd)
-                except Exception:
-                    print(
-                        f"{__title__}: error decompiling '{mainpyfile}'; disassembling instead.",
-                        file=sys.stderr,
+                if not no_decompiler:
+                    try:
+                        decompile_file(mainpyfile, fd.file, mapstream=fd)
+                        linemaps = decompile_file(mainpyfile, fd.file, mapstream=fd)
+                    except Exception:
+                        no_decompiler = True
+                        print(
+                            f"{__title__}: error decompiling '{mainpyfile}'; disassembling instead.",
+                            file=sys.stderr,
+                        )
+
+                if no_decompiler:
+                    info = disassemble_file(
+                        mainpyfile,
+                        outstream=fd,
+                        asm_format="extended-bytes",
+                        show_source=False,
                     )
-                    info = disassemble_file(mainpyfile, outstream=fd, asm_format="extended-bytes", show_source=False)
                     code_module = info[1]
-                    embeded_filename = code_module.co_filename
+                    embedded_filename = code_module.co_filename
 
                     old_tempfile = fd.name
                     fd.close()
@@ -205,18 +236,31 @@ def main(dbg=None, sys_argv=list(sys.argv)):
                         else:
                             print(
                                 "%s: couldn't find Python source '%s' or decompile it, so we disassembled it at '%s'"
-                                 % (__title__, embeded_filename, pyasm_name),
-                                 file=sys.stderr,
+                                % (__title__, embedded_filename, pyasm_name),
+                                file=sys.stderr,
                             )
-                            pyficache.remap_file(pyasm_name, embeded_filename)
+                            pyficache.remap_file(pyasm_name, embedded_filename)
 
                 else:
+                    decompile_file = fd.name
                     print(
-                        "%s: couldn't find Python source so we recreated it at '%s'"
-                        % (__title__, mainpyfile),
+                        "%s: couldn't find Python source, so we recreated it at '%s'."
+                        % (__title__, decompile_file),
                         file=sys.stderr,
                     )
-
+                    fd.close()
+                    embedded_filename = co.co_filename
+                    pyficache.remap_file(decompile_file, embedded_filename)
+                    pyficache.remap_file_lines(
+                        embedded_filename,
+                        decompile_file,
+                        (
+                            (orig_lineno, mapped_lineno)
+                            for orig_lineno, mapped_lineno in linemaps[
+                                0
+                            ].source_linemap.items()
+                        ),
+                    )
                 pass
 
         # If mainpyfile is an optimized Python script try to find and
